@@ -1,11 +1,11 @@
 ;;; org-modern.el --- Modern looks for Org -*- lexical-binding: t -*-
 
-;; Copyright (C) 2022  Free Software Foundation, Inc.
+;; Copyright (C) 2022-2023 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 0.6
+;; Version: 0.8
 ;; Package-Requires: ((emacs "27.1"))
 ;; Homepage: https://github.com/minad/org-modern
 
@@ -27,7 +27,7 @@
 ;;; Commentary:
 
 ;; This package adds some styling to your Org buffer, which gives it a
-;; modern look. Enable the styling by default with:
+;; modern look.  Enable the styling by default with:
 ;;   (add-hook 'org-mode-hook 'org-modern-mode)
 
 ;;; Code:
@@ -39,6 +39,9 @@
 
 (defgroup org-modern nil
   "Modern looks for Org."
+  :link '(info-link :tag "Info Manual" "(org-modern)")
+  :link '(url-link :tag "Homepage" "https://github.com/minad/org-modern")
+  :link '(emacs-library-link :tag "Library Source" "org-modern.el")
   :group 'org
   :prefix "org-modern-")
 
@@ -65,10 +68,11 @@ Set to nil to disable."
 
 (defcustom org-modern-timestamp t
   "Prettify time stamps, e.g. <2022-03-01>.
-Set to nil to disable styling the time stamps. In order to use custom
-timestamps, the format should be (DATE . TIME) where DATE is the format
-for date, and TIME is the format for time. DATE and TIME must be
-surrounded with space. For the syntax, refer to `format-time-string'."
+Set to nil to disable styling the time stamps.  In order to use
+custom timestamps, the format should be (DATE . TIME) where DATE
+is the format for date, and TIME is the format for time.  DATE
+and TIME must be surrounded with space.  For the syntax, refer to
+`format-time-string'."
   :type '(choice
           (const :tag "Disable time stamp styling" nil)
           (const :tag "Enable timestamp styling" t)
@@ -89,8 +93,19 @@ Set to nil to hide the vertical lines."
   :type '(choice (const nil) number))
 
 (defcustom org-modern-priority t
-  "Prettify priorities."
-  :type 'boolean)
+  "Prettify priorities.
+If set to t, the priority will be prettified with the brackets
+hidden.  If set to an alist of characters and strings, the
+associated string will be used as replacement for the given
+priority.  Example:
+
+  (setq org-modern-priority
+    (quote ((?A . \"❗\")
+            (?B . \"⬆\")
+            (?C . \"⬇\"))))"
+  :type '(choice (boolean :tag "Prettify")
+                 (alist :key-type (character :tag "Priority")
+                        :value-type (string :tag "Replacement"))))
 
 (defcustom org-modern-list
   '((?+ . "◦")
@@ -121,14 +136,30 @@ replacement expression, e.g., a string."
 (defcustom org-modern-todo-faces nil
   "Faces for todo keywords.
 This is an alist, with todo keywords in the car
-and faces in the cdr. Example:
+and faces in the cdr.  Example:
 
   (setq org-modern-todo-faces
     (quote ((\"TODO\" :background \"red\"
                     :foreground \"yellow\"))))"
   :type '(repeat
-          (cons (string :tag "Keyword")
-                (sexp   :tag "Face   "))))
+          (cons (choice
+                  (string :tag "Keyword")
+                  (const :tag "Default" t))
+                (sexp :tag "Face   "))))
+
+(defcustom org-modern-priority-faces nil
+  "Faces for priority tags.
+This is an alist, with priority character in the car and faces in
+the cdr.  Example:
+
+  (setq org-modern-priority-faces
+    (quote ((?A :background \"red\"
+                :foreground \"yellow\"))))"
+  :type '(repeat
+          (cons (choice
+                 (character :tag "Priority")
+                 (const :tag "Default" t))
+                (sexp :tag "Face   "))))
 
 (defcustom org-modern-tag t
   "Prettify tags in headlines, e.g., :tag1:tag2:."
@@ -157,16 +188,24 @@ all other blocks."
                         (string :tag "#+end_NAME replacement"))
                   (const :tag "Hide #+begin_ and #+end_ prefixes" t)))))
 
-(defcustom org-modern-block-fringe t
-  "Add a bitmap fringe to blocks."
-  :type 'boolean)
+(defcustom org-modern-block-fringe 0
+  "Add a border to the blocks in the fringe.
+This variable can also be set to an integer between 0 and 16,
+which specifies the offset of the block border from the edge of
+the window."
+  :type '(choice boolean integer))
 
 (defcustom org-modern-keyword t
   "Prettify keywords like #+title.
-If set to t, the prefix #+ will be hidden.
-If set to a string, e.g., \"‣\", the string is used as replacement for #+.
-If set to an alist of keywords and strings, the associated string will be
-used as replacement for \"#+keyword:\", with t the default key."
+If set to t, the prefix #+ will be hidden.  If set to a string,
+e.g., \"‣\", the string is used as replacement for #+.  If set to
+an alist of keywords and strings, the associated string will be
+used as replacement for \"#+keyword:\", with t the default key.
+Example:
+
+  (setq org-modern-keyword
+    (quote ((\"options\" . \"🔧\")
+            (t . t))))"
   :type '(choice (boolean :tag "Hide prefix")
                  (string :tag "Replacement")
                  (const :tag "Triangle bullet" "‣")
@@ -206,7 +245,7 @@ Set to nil to disable the indicator."
 
 (defface org-modern-symbol nil
   "Face used for stars, checkboxes and progress indicators.
-You can specify a font `:family'. The font families `Iosevka', `Hack' and
+You can specify a font `:family'.  The font families `Iosevka', `Hack' and
 `DejaVu Sans' give decent results.")
 
 (defface org-modern-label
@@ -214,7 +253,7 @@ You can specify a font `:family'. The font families `Iosevka', `Hack' and
   "Parent face for labels.
 The parameters of this face depend on typographical properties of
 the font and should therefore be adjusted by the user depending
-on their font, e.g., the :width or :height parameters. Themes
+on their font, e.g., the :width or :height parameters.  Themes
 should not override this face, since themes usually don't control
 the font.")
 
@@ -246,13 +285,13 @@ the font.")
   ;; `:inverse-video' to use todo foreground as label background
   '((t :inherit (org-todo org-modern-label)
        :weight semibold :inverse-video t))
-  "Face used for todo labels.")
+  "Default face used for todo labels.")
 
 (defface org-modern-priority
   ;; `:inverse-video' to use priority foreground as label background
   '((t :inherit (org-priority org-modern-label)
        :weight semibold :inverse-video t))
-  "Face used for priority labels.")
+  "Default face used for priority labels.")
 
 (defface org-modern-statistics
   '((t :inherit org-modern-done))
@@ -306,7 +345,7 @@ the font.")
         (end (match-end 1)))
     (put-text-property
      beg end 'display
-     (alist-get (char-after (1+ beg)) org-modern--checkbox-cache))))
+     (cdr (assq (char-after (1+ beg)) org-modern--checkbox-cache)))))
 
 (defun org-modern--keyword ()
   "Prettify keywords according to `org-modern-keyword'."
@@ -317,10 +356,27 @@ the font.")
     (unless rep
       (setq rep (cdr (assq t org-modern-keyword)) end (match-end 1)))
     (pcase rep
-      ('t (put-text-property beg (match-end 1) 'invisible t))
+      ('t (put-text-property beg (match-end 1) 'invisible 'org-modern))
       ((pred stringp)
        (put-text-property beg end 'display rep)))))
 
+(defun org-modern--priority ()
+  "Prettify priorities according to `org-modern-priority'."
+  (let* ((beg (match-beginning 1))
+         (end (match-end 1))
+         (prio (char-before (1- end))))
+    (if-let ((rep (and (consp org-modern-priority)
+                       (cdr (assq prio org-modern-priority)))))
+        (put-text-property beg end 'display rep)
+      (put-text-property beg (1+ beg) 'display " ")
+      (put-text-property (1- end) end 'display " ")
+      (put-text-property
+       beg end 'face
+       (if-let ((face (or (cdr (assq prio org-modern-priority-faces))
+                          (cdr (assq t org-modern-priority-faces)))))
+           `(:inherit (,face org-modern-label))
+         'org-modern-priority)))))
+
 (defun org-modern--progress ()
   "Prettify headline todo progress."
   (put-text-property
@@ -366,7 +422,8 @@ the font.")
     (put-text-property (1- end) end 'display (string (char-before end) ?\s))
     (put-text-property
      beg end 'face
-     (if-let (face (cdr (assoc todo org-modern-todo-faces)))
+     (if-let ((face (or (cdr (assoc todo org-modern-todo-faces))
+                        (cdr (assq t org-modern-todo-faces)))))
          `(:inherit (,face org-modern-label))
        (if (member todo org-done-keywords)
            'org-modern-done
@@ -423,10 +480,9 @@ the font.")
                                   (nth (logand i 1)
                                        org-modern--hide-stars-cache))))
     (when org-modern-star
-      (put-text-property
-       (match-beginning 2) (match-end 2) 'display
-       (aref org-modern--star-cache
-             (min (1- (length org-modern--star-cache)) level))))))
+      (put-text-property end (1+ end) 'display
+                         (aref org-modern--star-cache
+                               (min (1- (length org-modern--star-cache)) level))))))
 
 (defun org-modern--table ()
   "Prettify vertical table lines."
@@ -485,12 +541,12 @@ the font.")
     (pcase rep
       ('t
        (add-face-text-property beg-name end-name 'org-modern-block-name)
-       (put-text-property (if fringe beg-ind beg-rep) beg-name 'invisible t))
+       (put-text-property (if fringe beg-ind beg-rep) beg-name 'invisible 'org-modern))
       ((pred stringp)
        (add-face-text-property beg-name end-name 'org-modern-block-name)
        (put-text-property beg-rep end-rep 'display rep)
        (when fringe
-         (put-text-property beg-ind beg-rep 'invisible t))))))
+         (put-text-property beg-ind beg-rep 'invisible 'org-modern))))))
 
 (defun org-modern--block-fringe ()
   "Prettify blocks with fringe bitmaps."
@@ -529,8 +585,8 @@ the font.")
 
 (defun org-modern--pre-redisplay (_)
   "Compute font parameters before redisplay."
-  (when-let (box (and org-modern-label-border
-                      (face-attribute 'org-modern-label :box nil t)))
+  (when-let ((box (and org-modern-label-border
+                       (face-attribute 'org-modern-label :box nil t))))
     (unless (equal (and (listp box) (plist-get box :color))
                    (face-attribute 'default :background nil t))
       (org-modern--update-label-face)))
@@ -554,7 +610,7 @@ the font.")
        (list :color (face-attribute 'default :background nil t)
              :line-width
              ;; Emacs 28 supports different line horizontal and vertical line widths
-             (if (>= emacs-major-version 28)
+             (if (eval-when-compile (>= emacs-major-version 28))
                  (cons 0 (- border))
                (- border)))))))
 
@@ -564,28 +620,35 @@ the font.")
              (fboundp 'fringe-bitmap-p)
              (not (fringe-bitmap-p 'org-modern--block-inner)))
     (let* ((g (ceiling (frame-char-height) 1.8))
-           (h (- (default-line-height) g)))
+           (h (- (default-line-height) g))
+           (v (expt 2 (- 15 (if (booleanp org-modern-block-fringe) 0
+                              org-modern-block-fringe))))
+           (w (+ v v -1)))
       (define-fringe-bitmap 'org-modern--block-inner
-        [128] nil nil '(top t))
+        (vector v) nil 16 '(top t))
       (define-fringe-bitmap 'org-modern--block-begin
-        (vconcat (make-vector g 0) [#xFF] (make-vector (- 127 g) #x80)) nil nil 'top)
+        (vconcat (make-vector g 0) (vector w) (make-vector (- 127 g) v)) nil 16 'top)
       (define-fringe-bitmap 'org-modern--block-end
-        (vconcat (make-vector (- 127 h) #x80) [#xFF] (make-vector h 0)) nil nil 'bottom))))
+        (vconcat (make-vector (- 127 h) v) (vector w) (make-vector h 0)) nil 16 'bottom))))
+
+(defun org-modern--symbol (str)
+  "Add `org-modern-symbol' face to STR."
+  (setq str (copy-sequence str))
+  (add-face-text-property 0 (length str) 'org-modern-symbol 'append str)
+  str)
 
 (defun org-modern--make-font-lock-keywords ()
   "Compute font-lock keywords."
   (append
-   (when-let (bullet (alist-get ?+ org-modern-list))
+   (when-let ((bullet (alist-get ?+ org-modern-list)))
      `(("^[ \t]*\\(+\\)[ \t]" 1 '(face nil display ,bullet))))
-   (when-let (bullet (alist-get ?- org-modern-list))
+   (when-let ((bullet (alist-get ?- org-modern-list)))
      `(("^[ \t]*\\(-\\)[ \t]" 1 '(face nil display ,bullet))))
-   (when-let (bullet (alist-get ?* org-modern-list))
+   (when-let ((bullet (alist-get ?* org-modern-list)))
      `(("^[ \t]+\\(*\\)[ \t]" 1 '(face nil display ,bullet))))
    (when org-modern-priority
-     '(("^\\*+.*? \\(\\(\\[\\)#.\\(\\]\\)\\) "
-        (1 'org-modern-priority t)
-        (2 '(face nil display " "))
-        (3 '(face nil display " ")))))
+     `(("^\\*+.*? \\(\\(\\[\\)#.\\(\\]\\)\\) "
+        (1 (org-modern--priority)))))
    (when org-modern-todo
      `((,(format "^\\*+ +%s " (regexp-opt org-todo-keywords-1 t))
         (0 (org-modern--todo)))))
@@ -593,12 +656,12 @@ the font.")
      '(("^[ \t]*\\(?:[-+*]\\|[0-9]+[.)]\\)[ \t]+\\(\\[[ X-]\\]\\)[ \t]"
         (0 (org-modern--checkbox)))))
    (when (or org-modern-star org-modern-hide-stars)
-     `(("^\\(\\**\\)\\(\\*\\) "
+     `(("^\\(\\**\\)\\* "
         ,@(and (not (eq org-modern-hide-stars t))
                (or org-modern-star (stringp org-modern-hide-stars))
                '((0 (org-modern--star))))
-        ,@(and (eq org-modern-hide-stars 'leading) '((1 '(face nil invisible t))))
-        ,@(and (eq org-modern-hide-stars t) '((0 '(face nil invisible t)))))))
+        ,@(and (eq org-modern-hide-stars 'leading) '((1 '(face nil invisible org-modern))))
+        ,@(and (eq org-modern-hide-stars t) '((0 '(face nil invisible org-modern)))))))
    (when org-modern-horizontal-rule
      `(("^[ \t]*-\\{5,\\}$" 0
         '(face org-modern-horizontal-rule display
@@ -612,12 +675,12 @@ the font.")
         (0 (org-modern--tag)))))
    (when org-modern-footnote
      `(("^\\(\\[fn:\\)[[:word:]-_]+\\]" ;; Definition
-        ,@(if-let (x (car org-modern-footnote))
+        ,@(if-let ((x (car org-modern-footnote)))
               `((0 '(face nil display ,x))
                 (1 '(face nil display ,(propertize "[" 'display x))))
             '((1 '(face nil display "[")))))
        ("[^\n]\\(\\(\\[fn:\\)[[:word:]-_]+\\]\\)" ;; Reference
-        ,@(if-let (x (cdr org-modern-footnote))
+        ,@(if-let ((x (cdr org-modern-footnote)))
               `((1 '(face nil display ,x))
                 (2 '(face nil display ,(propertize "[" 'display x))))
             '((2 '(face nil display "[")))))))
@@ -626,21 +689,17 @@ the font.")
       (when org-modern-internal-target
         `((,(format "\\(<<\\)%s\\(>>\\)" target)
            (0 '(face org-modern-internal-target) t)
-           (1 '(face nil display ,(propertize (car org-modern-internal-target)
-                                              'face 'org-modern-symbol)))
-           (3 '(face nil display ,(propertize (caddr org-modern-internal-target)
-                                              'face 'org-modern-symbol)))
+           (1 '(face nil display ,(org-modern--symbol (car org-modern-internal-target))))
+           (3 '(face nil display ,(org-modern--symbol (caddr org-modern-internal-target))))
            ,@(unless (cadr org-modern-internal-target)
-               '((2 '(face nil invisible t)))))))
+               '((2 '(face nil invisible org-modern)))))))
       (when org-modern-radio-target
         `((,(format "\\(<<<\\)%s\\(>>>\\)" target)
            (0 '(face org-modern-radio-target) t)
-           (1 '(face nil display ,(propertize (car org-modern-radio-target)
-                                              'face 'org-modern-symbol)))
-           (3 '(face nil display ,(propertize (caddr org-modern-radio-target)
-                                              'face 'org-modern-symbol)))
+           (1 '(face nil display ,(org-modern--symbol (car org-modern-radio-target))))
+           (3 '(face nil display ,(org-modern--symbol (caddr org-modern-radio-target))))
            ,@(unless (cadr org-modern-radio-target)
-               '((2 '(face nil invisible t)))))))))
+               '((2 '(face nil invisible org-modern)))))))))
    (when org-modern-timestamp
      '(("\\(?:<\\|\\[\\)\\([0-9]\\{4\\}-[0-9]\\{2\\}-[0-9]\\{2\\}\\(?: [[:word:]]+\\.?\\)?\\(?: [.+-]+[0-9ymwdh/]+\\)*\\)\\(\\(?: [0-9:-]+\\)?\\(?: [.+-]+[0-9ymwdh/]+\\)*\\)\\(?:>\\|\\]\\)"
         (0 (org-modern--timestamp)))
@@ -656,7 +715,7 @@ the font.")
    (when org-modern-keyword
      `(("^[ \t]*\\(#\\+\\)\\([^: \t\n]+\\):"
         ,@(pcase org-modern-keyword
-            ('t '(1 '(face nil invisible t)))
+            ('t '(1 '(face nil invisible org-modern)))
             ((pred stringp) `(1 '(face nil display ,org-modern-keyword)))
             (_ '(0 (org-modern--keyword)))))))
    ;; Do not add source block fringe markers if org-indent-mode is
@@ -668,9 +727,9 @@ the font.")
    (when org-modern-block-name
      (let* ((indent (and org-modern-block-fringe
                          (not (bound-and-true-p org-indent-mode))
-                         '((1 '(face nil invisible t)))))
+                         '((1 '(face nil invisible org-modern)))))
             (name '(3 'org-modern-block-name append))
-            (hide `(,@indent (2 '(face nil invisible t)) ,name))
+            (hide `(,@indent (2 '(face nil invisible org-modern)) ,name))
             (specs
              (pcase org-modern-block-name
                ('t ;; Hide
@@ -692,22 +751,20 @@ the font.")
   :group 'org-modern
   (cond
    (org-modern-mode
+    (add-to-invisibility-spec 'org-modern)
     (setq
      org-modern--star-cache
-     (vconcat (mapcar
-               (lambda (x) (propertize x 'face 'org-modern-symbol))
-               org-modern-star))
+     (vconcat (mapcar #'org-modern--symbol org-modern-star))
      org-modern--hide-stars-cache
      (and (stringp org-modern-hide-stars)
-          (list (propertize org-modern-hide-stars 'face 'org-modern-symbol)
-                (propertize org-modern-hide-stars 'face 'org-modern-symbol)))
+          (list (org-modern--symbol org-modern-hide-stars)
+                (org-modern--symbol org-modern-hide-stars)))
      org-modern--progress-cache
      (vconcat (mapcar
-               (lambda (x) (concat " " (propertize x 'face 'org-modern-symbol) " "))
+               (lambda (x) (concat " " (org-modern--symbol x) " "))
                org-modern-progress))
      org-modern--checkbox-cache
-     (mapcar (pcase-lambda (`(,k . ,v))
-               (cons k (propertize v 'face 'org-modern-symbol)))
+     (mapcar (pcase-lambda (`(,k . ,v)) (cons k (org-modern--symbol v)))
              org-modern-checkbox)
      org-modern--font-lock-keywords
      (append (remove '(org-fontify-meta-lines-and-blocks) org-font-lock-keywords)
@@ -719,6 +776,7 @@ the font.")
     (org-modern--update-label-face)
     (org-modern--update-fringe-bitmaps))
    (t
+    (remove-from-invisibility-spec 'org-modern)
     (font-lock-remove-keywords nil org-modern--font-lock-keywords)
     (font-lock-add-keywords nil org-font-lock-keywords)
     (setq-local font-lock-unfontify-region-function #'org-unfontify-region)
@@ -743,6 +801,8 @@ the font.")
 ;;;###autoload
 (defun org-modern-agenda ()
   "Finalize Org agenda highlighting."
+  (remove-from-invisibility-spec 'org-modern)
+  (add-to-invisibility-spec 'org-modern) ;; Not idempotent?!
   (add-hook 'pre-redisplay-functions #'org-modern--pre-redisplay nil 'local)
   (save-excursion
     (save-match-data
@@ -762,12 +822,11 @@ the font.")
             (org-modern--tag))))
       (when org-modern-priority
         (goto-char (point-min))
-        (while (re-search-forward "\\(\\[\\)#.\\(\\]\\)" nil 'noerror)
+        (while (re-search-forward "\\(\\[#.\\]\\)" nil 'noerror)
           ;; For some reason the org-agenda-fontify-priorities adds overlays?!
-          (when-let (ov (overlays-at (match-beginning 0))) (overlay-put (car ov) 'face nil))
-          (put-text-property (match-beginning 0) (match-end 0) 'face 'org-modern-priority)
-          (put-text-property (match-beginning 1) (match-end 1) 'display " ")
-          (put-text-property (match-beginning 2) (match-end 2) 'display " "))))))
+          (when-let ((ov (overlays-at (match-beginning 0))))
+            (overlay-put (car ov) 'face nil))
+          (org-modern--priority))))))
 
 ;;;###autoload
 (define-globalized-minor-mode global-org-modern-mode
